// Kernel bandwidth
const decl SKDE_BANDWIDTH = 0.5;

/*
	The GAS DGPs.

	Common arguments:
		eps_t: a 1xT vector of i.i.d. random variates from the desired density q(.) with zero mean and unit variance (i.e. rann(1, T))
		mu: mean
		omega, alpha, beta: GAS parameters

	Returns an array:
		[0]: resulting data
		[1]: underlying f_t
*/
GASgenerator_Normal(const T, const mu, const omega, const alpha, const beta) {
	decl eps_t = rann(1, T), f_t = zeros(1, T), y_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		y_t[i] = mu + exp(f_t[i]/2) * eps_t[i];
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlogn(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {y_t, f_t};
}

GASgenerator_t(const T, const mu, const omega, const alpha, const beta, const nu) {
	decl eps_t = rannormt(1, T, nu), f_t = zeros(1, T), y_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		y_t[i] = mu + exp(f_t[i]/2) * eps_t[i];
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlognormt(eps_t[i], nu) + 1) + beta * f_t[i];
		++i;
	}
	return {y_t, f_t};
}

GASgenerator_MN(const T, const mu, const omega, const alpha, const beta) {
	decl eps_t = ranchimix(1, T), f_t = zeros(1, T), y_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		y_t[i] = mu + exp(f_t[i]/2) * eps_t[i];
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlogchimix(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {y_t, f_t};
}

/*
	Non-GAS generators.

	Arguments:
		T: time series length

		Returns a 1xT vector of volatility
*/

Vol_Sine(const T) {
	return 2.5 + cos(2*M_PI*range(0, T-1)/200);
}

Vol_LongRegimes(const T) {
	return 1.5 + 2 * (1 - imod(range(0, T-1)/500, 2));
}

Vol_ShortRegimes(const T) {
	return 1.5 + 2 * (1 - imod(range(0, T-1)/200, 2));
}

Vol_Bursts(const T) {
	decl result = <>;
	decl onecycle = 1.5 * ones(1, 150) ~ range(1.9, 3.5, 0.4) ~ range(3.5-2/95, 1.5, -2/95);
	decl bursts = idiv(T, 250);
	for(decl i = 0; i < bursts; i++) {
		result ~= onecycle;
	}
	result ~= 1.5 * ones(1, T - bursts * 250);
	return result;
}

Vol_Stoch(const T) {
	decl alpha = 0.5;
	decl eta = 0.1 * rann(1, T);
	decl result = zeros(1, T);
	for(decl i = 0; i < T; i++) {
		alpha =	0.01 + 0.98 * alpha + eta[i];
		result[i] = exp(alpha/2);
	}
	return result;
}

/*
	The GAS(1,1) filters.

	Common arguments:
		data: a 1xT vector of data
		mu: mean
		omega, alpha, beta: GAS parameters

	Returns an array with the following elements:
	[0]: filtered f_t
	[1]: filtered standardized residuals
*/

GASfilter_Normal(const data, const mu, const omega, const alpha, const beta) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlogn(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

// extra argument: nu
GASfilter_t(const data, const mu, const omega, const alpha, const beta, const nu) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlognormt(eps_t[i], nu) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

// extra argument: StandardizedKernelEstimator object
GASfilter_SP(const data, const mu, const omega, const alpha, const beta, const skde) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * skde.dlog(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

GASfilter_MN(const data, const mu, const omega, const alpha, const beta) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlogchimix(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

/*
	The GARCH(1,1) filter.
*/
GARCHfilter(const data, const mu, const omega, const alpha, const beta) {
	decl T = columns(data), sigma2_t = zeros(1, T), xi_t = zeros(1, T);
	sigma2_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		xi_t[i] = data[i] - mu;
		if(i == T-1) break;
		sigma2_t[i+1] = omega * (1 - beta - alpha) + alpha * xi_t[i]^2 + beta * sigma2_t[i];
		++i;
	}
	return {sigma2_t, xi_t./sqrt(sigma2_t)};
}

/*
	Estimators. All estimators have the following form:

	Arguments:
		data: a 1xT vector of data
		K: how much of the data to use for estimation
		apar: an adress, holds a vector with first elements mu, omega, alpha, beta (some estimators include additional parameters)
			in: starting parameters for optimization
			out: optimized parameters
		aft: an address
			out: filtered f_t (including for times that were excluded in the optimization)
		aepst: an address
			out: filtered standardized residuals

	Returns:
		The MaxBFGS return code
*/

GAS_QMLE(const data, const K, const apar, const aft, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 0, 0, 0>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, 0, 0, 0>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_Normal(data[0:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(densn(eps_t))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, 0, 0, 0>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[aft[0], aepst[0]] = GASfilter_Normal(data, mu, omega, alpha, beta);
	
	return code;	
}

GAS_t(const data, const K, const apar, const aft, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 0, 0, 0, 1>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, 0, 0, 0, -1>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_t(data[0:K-1], mu, omega, alpha, beta, nu);
			aval[0] = meanr(log(densnormt(eps_t, nu))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, 0, 0, 0, -1>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
	[aft[0], aepst[0]] = GASfilter_t(data, mu, omega, alpha, beta, nu);
	
	return code;	
}

GAS_SP1(const data, const K, const apar, const aft, const aepst) {
	decl tft, tepst, tpar;
	tpar = apar[0] ~ <5>;
	decl tcode = GAS_t(data, K, &tpar, &tft, &tepst);
	if (tcode > 0) return tcode;
	
	decl skde = new StandardizedKernelEstimator(tepst[0:K-1], SKDE_BANDWIDTH);

	decl par = apar[0];
	par = Transform(par, <0, 0, 0, 0>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, 0, 0, 0>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_SP(data[0:K-1], mu, omega, alpha, beta, skde);
			aval[0] = meanr(log(skde.dens(eps_t))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, 0, 0, 0>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[aft[0], aepst[0]] = GASfilter_SP(data, mu, omega, alpha, beta, skde);
	
	return code;
}

/*
	Specific MLE's
*/

GAS_t3(const data, const K, const apar, const aft, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 0, 0, 0>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, 0, 0, 0>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_t(data[0:K-1], mu, omega, alpha, beta, 3);
			aval[0] = meanr(log(densnormt(eps_t, 3))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, 0, 0, 0>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[aft[0], aepst[0]] = GASfilter_t(data, mu, omega, alpha, beta, 3);
	
	return code;	
}

GAS_t5(const data, const K, const apar, const aft, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 0, 0, 0>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, 0, 0, 0>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_t(data[0:K-1], mu, omega, alpha, beta, 5);
			aval[0] = meanr(log(densnormt(eps_t, 5))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, 0, 0, 0>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[aft[0], aepst[0]] = GASfilter_t(data, mu, omega, alpha, beta, 5);
	
	return code;	
}

GAS_MN(const data, const K, const apar, const aft, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 0, 0, 0>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, 0, 0, 0>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_MN(data[0:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(denschimix(eps_t))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, 0, 0, 0>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[aft[0], aepst[0]] = GASfilter_MN(data, mu, omega, alpha, beta);
	
	return code;	
}

/*
	GARCH Estimators
*/

GARCH_QMLE(const data, const K, const apar, const asigma2t, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 1, 0, 2>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, -1, 0, -2>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl sigma2_t, eps_t;
			[sigma2_t, eps_t] = GARCHfilter(data[0:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(densn(eps_t))) - meanr(0.5*log(sigma2_t));
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, -1, 0, -2>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[asigma2t[0], aepst[0]] = GARCHfilter(data, mu, omega, alpha, beta);
	
	return code;
}

GARCH_t(const data, const K, const apar, const asigma2t, const aepst) {
	decl par = apar[0];
	par = Transform(par, <0, 1, 0, 2, 1>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, -1, 0, -2, -1>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
			
			decl sigma2_t, eps_t;
			[sigma2_t, eps_t] = GARCHfilter(data[0:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(densnormt(eps_t, nu))) - meanr(0.5*log(sigma2_t));
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, -1, 0, -2, -1>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
	[asigma2t[0], aepst[0]] = GARCHfilter(data, mu, omega, alpha, beta);
	
	return code;	
}

GARCH_SP1(const data, const K, const apar, const asigma2t, const aepst) {
	decl tft, tepst, tpar;
	tpar = apar[0] ~ <5>;
	decl tcode = GARCH_t(data, K, &tpar, &tft, &tepst);
	if (tcode > 0) return tcode;
	
	decl skde = new StandardizedKernelEstimator(tepst[0:K-1], SKDE_BANDWIDTH);

	decl par = apar[0];
	par = Transform(par, <0, 1, 0, 2>)';
	decl val;

	decl code = MaxBFGS(
		[=] (par, const aval, const a2, const a3) {
			par = Transform(par, <0, -1, 0, -2>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl sigma2_t, eps_t;
			[sigma2_t, eps_t] = GARCHfilter(data[0:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(skde.dens(eps_t))) - meanr(0.5*log(sigma2_t));
			
			return 1;
		}
	, &par, &val, 0, 1);

	par = Transform(par, <0, -1, 0, -2>)';
	apar[0] = par;
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	[asigma2t[0], aepst[0]] = GARCHfilter(data, mu, omega, alpha, beta);
	
	return code;
}