/*
	The kernel estimation class.

	A kernel density estimate is constructed with the given points and bandwidth (and the Gaussian kernel).
	It is automatically normalized to zero mean and unit variance.
*/
#include <oxstd.h>
#include <oxfloat.h>
static decl KERNEL_T_NU = 5;
static decl KERNEL_T_INTEGRATING_CONSTANT = 1;



class StandardizedKernelEstimator {
	decl points, bandwidth;
	StandardizedKernelEstimator(const points, const bandwidth);
	dens(const x);
	ddens(const x);
	dlog(const x);
};

/*
	Construct the standardized kernel estimate.

	points: 1xT vector of data points
	bandwidth: pre-normalization bandwidth
*/
StandardizedKernelEstimator::StandardizedKernelEstimator(const points, const bandwidth) {
	decl mixturemean = meanr(points);
	decl mixturesd = sqrt(meanr((points-mixturemean).^2 + bandwidth^2));
	this.points = (points - mixturemean) / mixturesd;
	this.bandwidth = bandwidth / mixturesd;
}

/*
	Compute the density estimate.

	x: 1xn vector of values at which to evaluate the density
*/
StandardizedKernelEstimator::dens(const x) {
	KERNEL_T_INTEGRATING_CONSTANT = exp( loggamma((KERNEL_T_NU+1)/2) - loggamma(KERNEL_T_NU/2) - 0.5*log((KERNEL_T_NU-2)*M_PI));
	return meanr(
		((1 + sqr((x' - points) / bandwidth)/(KERNEL_T_NU-2)).^(-0.5*(KERNEL_T_NU+1)))
	)' * KERNEL_T_INTEGRATING_CONSTANT / bandwidth;
}

/*
	Compute the derivative of the density.

	x: 1xn vector values at which to evaluate the derivative
*/
StandardizedKernelEstimator::ddens(const x) {
	KERNEL_T_INTEGRATING_CONSTANT = exp( loggamma((KERNEL_T_NU+1)/2) - loggamma(KERNEL_T_NU/2) - 0.5*log((KERNEL_T_NU-2)*M_PI));
	return -(KERNEL_T_NU+1)/(KERNEL_T_NU-2) * meanr(
		(x' - points)
		.* 
		((1 + sqr((x' - points) / bandwidth)/(KERNEL_T_NU-2)).^(-0.5*(KERNEL_T_NU+3)))
	)' *  KERNEL_T_INTEGRATING_CONSTANT / bandwidth^3;
}

/*
	Compute the derivative of the logarithm of the density (used for SP-GAS).

	x: 1xn vector values at which to evaluate the derivative
*/
StandardizedKernelEstimator::dlog(const x) {
	return ddens(x) ./ dens(x);
}